#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <tkplayer.h>
#include <tkiocomm.h>

#define BUFSIZE 8192L

void copybytes(FILE *srchandle, FILE *desthandle, fpos_t bytes)
{ // assumes positions already set!
  long tocopy,count;
  char *buffer;

  if (bytes == 0) return;

  fflush(srchandle);
  fflush(desthandle);

  tocopy=bytes;
  count=bytes;
  if (count > BUFSIZE) count=BUFSIZE;

  buffer=(char *)malloc(count);
  if (!buffer) { // after all my hard work, too...
    return;
  }
  do {
    fread(buffer,1,count,srchandle);
    fwrite(buffer,1,count,desthandle);
    tocopy -= count;
    if (tocopy < BUFSIZE) count=tocopy;
  } while (tocopy > 0);
  free(buffer);

  fflush(srchandle);
  fflush(desthandle);
}


int main(int argc, char *argv[]) {
  int c;
  unsigned long toplayers=0;
  XString dir;
  XString file;
  FILE *fi,*fo;
  bool atstart=true;

  opterr = 0;
  while ((c=getopt(argc, argv, "123456789aAbBbd:f:hse")) != -1) {
    switch (c) {
      case '1':
        toplayers |= 0x1;
        break;
      case '2':
        toplayers |= 0x2;
        break;
      case '3':
        toplayers |= 0x4;
        break;
      case '4':
        toplayers |= 0x8;
        break;
      case '5':
        toplayers |= 0x10;
        break;
      case '6':
        toplayers |= 0x20;
        break;
      case '7':
        toplayers |= 0x40;
        break;
      case '8':
        toplayers |= 0x80;
        break;
      case '9':
        toplayers |= 0x100;
        break;
      case 'A':
      case 'a':
        toplayers |= 0x200;
        break;
      case 'B':
      case 'b':
        toplayers |= 0x400;
        break;
      case 'd':
        dir=optarg;
        break;
      case 'f':
        file=optarg;
        break;
      case 's':
        atstart=true;
        break;
      case 'e':
        atstart=false;
        break;
      case 'h':
        // help();
        exit(1);
      case '?':
        fprintf(stderr,"Unknown option %c - use -h for help\n", c);
        exit(1);
    }
  }

  if (!dir) {
    fprintf(stderr,"Specify valid directory with -d <dir>\n");
    exit(1);
  }

  if (!file) {
    fprintf(stderr,"Specify file with -f <file>\n");
    exit(1);
  }

  if (toplayers==0) {
    fprintf(stderr,"Specify players with -1...-B\n");
    exit(1);
  }

  for (int pl=1;pl<12;pl++) {
    if ((1<<(pl-1)) & toplayers) {
      XString fin,fon;
      XString winver="";
      char pf[20];
      long fol=0;
      sprintf(pf,"%d",pl);
#ifdef DEBUG
      fin=dir | "\\player";
      fin |= pf;
      fin |= ".rst";
      fon=dir | "\\player";
      fon |= pf;
      fon |= ".new";
#else
      fin=dir | "\\player";
      fin |= pf;
      fin |= ".bak";
      fon=dir | "\\player";
      fon |= pf;
      fon |= ".rst";
      remove((const char *)fin);
      rename((const char *)fon,(const char *)fin);
#endif
      fi=fopen((const char *)fin,"rb");
      fo=fopen((const char *)fon,"wb+");
      if (fi && fo) {
        char *x;
        UB4 fip,fop;

        // Start new RST; we'll fill this in later
        x=(char *)calloc(96,1);
        if (x) {
          fwrite(x,1,96,fo);
          fol=96;
          free(x);
        } else {
          fprintf(stderr,"Panic!\n");
        }

        // copy the standard 8 (except messages)
        for (int k=1;k<9;k++) {
          int mode=0,size=0,extent=0;
          fseek(fi,(k-1)*4,SEEK_SET);
          fip=readUB4(fi);
          switch (k) {
            case 1:
              size=107;
              mode=1;
              break;
            case 2:
              size=34;
              mode=1;
              break;
            case 3:
              size=85;
              mode=1;
              break;
            case 4:
              size=156;
              mode=1;
              break;
            case 5:
              break;
            case 6:
              size=4000;
              mode=2;
              break;
            case 7:
              size=144;
              mode=2;
              break;
            case 8:
              size=100;
              mode=1;
              break;
          };
          fseek(fo,fol,SEEK_SET);
          fop=ftell(fo);
          switch (mode) {
            case 1:
              fseek(fi,fip-1,SEEK_SET);
              extent=readUB2(fi)*size+2;
              fseek(fi,fip-1,SEEK_SET);
#ifdef DEBUG
              printf("Writing section %d, %d bytes\n",k,extent);
#endif
              copybytes(fi,fo,extent);
              fol+=extent;
              fseek(fo,(k-1)*4,SEEK_SET);
              writeUB4(fo,fop+1);
              break;
            case 2:
              fseek(fi,fip-1,SEEK_SET);
#ifdef DEBUG
              printf("Writing section %d, %d bytes\n",k,size);
#endif
              copybytes(fi,fo,size);
              fol+=size;
              fseek(fo,(k-1)*4,SEEK_SET);
              writeUB4(fo,fop+1);
              break;
          }
        }

        // check for WinPlan data, copy it
        fseek(fi,32,SEEK_SET);
        fseek(fo,32,SEEK_SET);
        if (readXString(fi,6)=="VER3.5") {
          int extent=0;
          winver="VER3.5" | readXString(fi,2);
          writeXString(fo,8,winver);
          // pure WinPlan part
          fip=readUB4(fi);
          fseek(fi,fip-1+13282,SEEK_SET);
          if (readXString(fi,4)=="1120") { // extended contacts database
            extent=13286+2+34*readUB2(fi);
          } else {
            extent=13286;
          }
          fseek(fo,fol,SEEK_SET);
          fop=ftell(fo);
          fseek(fi,fip-1,SEEK_SET);
#ifdef DEBUG
              printf("Writing WinPlan, %d bytes\n",extent);
#endif
          copybytes(fi,fo,extent);
          fol+=extent;
          fseek(fo,40,SEEK_SET);
          writeUB4(fo,fop+1);

          //leechX.dat
          fseek(fi,44,SEEK_SET);
          fip=readUB4(fi);
          if (fip!=0) {
            fseek(fo,fol,SEEK_SET);
            fop=ftell(fo);
            fseek(fi,fip-1,SEEK_SET);
            extent=readUB4(fi)+4;
            fseek(fi,fip-1,SEEK_SET);
#ifdef DEBUG
              printf("Writing leechX, %d bytes\n",extent);
#endif
            copybytes(fi,fo,extent);
            fol+=extent;
            fseek(fo,44,SEEK_SET);
            writeUB4(fo,fop+1);
          } else {
            fseek(fo,44,SEEK_SET);
            writeUB4(fo,0);
          }

          //extended UFO database
          fseek(fi,48,SEEK_SET);
          fip=readUB4(fi);
          if (fip!=0) {
            fseek(fo,fol,SEEK_SET);
            fop=ftell(fo);
            fseek(fi,fip-1,SEEK_SET);
            extent=(readUB2(fi)-100)*78+2;
            if (extent<0) extent=0;
            fseek(fi,fip-1,SEEK_SET);
#ifdef DEBUG
              printf("Writing x-UFO, %d bytes\n",extent);
#endif
            copybytes(fi,fo,extent);
            fol+=extent;
            fseek(fo,48,SEEK_SET);
            writeUB4(fo,fop+1);
          } else {
            fseek(fo,48,SEEK_SET);
            writeUB4(fo,0);
          }
        }

        // now for the tricky bit! The messages...
        fop=fol;
        fseek(fo,16,SEEK_SET);
        writeUB4(fo,fop+1);
        {
          INMESS messpad[500];
          int nmess,delta=0,nst;
          char *x;
          FILE *newmess;

          fseek(fi,16,SEEK_SET);
          fip=readUB4(fi);
          fseek(fi,fip-1,SEEK_SET);
          nmess=readUB2(fi);
          if (atstart) {
            nst=0;
          } else {
            nst=nmess;
          }

          newmess=fopen((const char *)file,"rt");
          if (newmess) {
            char inbuff[100];
            XString mip="",lip,buff="";
            int lc=0;
            bool eom=false;

            while (!feof(newmess)) {
              fgets(inbuff,99,newmess);
              lip=inbuff;
              if (feof(newmess)) {
                lip=".\n";
              }
              if (lip) {
                lip=lip.Left(lip.Len()-1);
                if (buff) {
                  lip=buff | '\r' | lip;
                  buff="";
                }
                while (lip.Len()>40) {
                  int j=40;
                  do {
                    j--;
                    if (lip.GetAt(j)==' ') {
                      lip.SetAt(j,'\r');
                      j=-1;
                    }
                  } while (j>=0);
                  j=lip.Find('\r');
                  if (j==-1) {
                    mip |= lip.Left(40) | '\r';
                    lip=lip.Mid(40);
                  } else {
                    mip |= lip.Left(j+1);
                    lip=lip.Mid(j+1);
                  }
                  lc++;
                  if (lc>18) {
                    buff=lip;
                    lip="";
                  }
                };
                if (lip==".") {
                  eom=true;
                } else {
                  mip |= lip | '\r';
                  lc++;
                  if (lc>19) {
                    eom=true;
                  }
                }
                if (eom) {
                  if (mip) {
                    messpad[nst].sMessage=mip;
                    nst++;
                  }
                  mip="";
                  lc=0;
                  eom=false;
                }
              }
            }
            fclose(newmess);
            if (atstart) {
              delta=nst;
            } else {
              delta=0;
            }
          }

          // read old
          for (int k=0;k<nmess;k++) {
            fseek(fi,fip-1+2+6*k,SEEK_SET);
            messpad[k+delta].lOffset=readUB4(fi);
            messpad[k+delta].wLength=readUB2(fi);
            fseek(fi,messpad[k+delta].lOffset-1,SEEK_SET);
            messpad[k+delta].sMessage=readXStringD(fi,messpad[k+delta].wLength);
          }

          // write out, afresh
          if (atstart) {
            nmess+=nst;
          } else {
            nmess=nst;
          }
          fseek(fo,fol,SEEK_SET);
          writeUB2(fo,nmess);
          fol+=2;
          fop=fol;
          x=(char *)calloc(1,nmess*6);
          if (x) {
            fwrite(x,1,nmess*6,fo);
            fol+=nmess*6;
            free(x);
          } else {
            fprintf(stderr,"Panic!\n");
          }
          for (int k=0;k<nmess;k++) {
            fseek(fo,fol,SEEK_SET);
            messpad[k].lOffset=fol+1;
            messpad[k].sMessage.Trim();
            writeXStringD(fo,messpad[k].sMessage.Len(),messpad[k].sMessage);
            fol+=messpad[k].sMessage.Len();
            fseek(fo,fop+k*6,SEEK_SET);
            writeUB4(fo,messpad[k].lOffset);
            writeUB2(fo,messpad[k].sMessage.Len());
          }
        }
        fclose(fi);
        fclose(fo);
        fprintf(stderr,"Wrote message to player%d.rst.\n",pl);
      } else {
        if (fi) fclose(fi);
        if (fo) fclose(fo);
        remove((const char *)fon);
        rename((const char *)fin,(const char *)fon);
        fprintf(stderr,"Can't open files for player%d.rst.\n",pl);
      }
    }
  }
}
